#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import copy
import json
import math
from typing import List

import flsim.configs  # noqa
import pkg_resources
import pytest
import torch
from flsim.active_user_selectors.simple_user_selector import (
    SequentialActiveUserSelectorConfig,
)
from flsim.clients.base_client import ClientConfig
from flsim.common.pytest_helper import (
    assertEqual,
    assertTrue,
    assertIsInstance,
    assertEmpty,
)
from flsim.common.timeout_simulator import GaussianTimeOutSimulatorConfig
from flsim.data.data_provider import FLDataProviderFromList
from flsim.data.data_sharder import SequentialSharder
from flsim.data.dataset_data_loader import FLDatasetDataLoaderWithBatch
from flsim.fb.data.hive_data_utils import create_dataloader
from flsim.fb.data.hive_dataset import InlineDatasetConfig
from flsim.fb.data.paged_dataloader import PagedDataProvider
from flsim.interfaces.metrics_reporter import TrainingStage
from flsim.optimizers.async_aggregators import FedAdamAsyncAggregatorConfig
from flsim.optimizers.local_optimizers import (
    LocalOptimizerSGDConfig,
    LocalOptimizerFedProxConfig,
)
from flsim.optimizers.optimizer_scheduler import ArmijoLineSearchSchedulerConfig
from flsim.optimizers.server_optimizers import (
    FedAdamOptimizerConfig,
    FedAvgWithLROptimizerConfig,
)
from flsim.servers.sync_servers import (
    SyncServerConfig,
)
from flsim.tests.utils import (
    FakeMetricReporter,
    MockRecord,
    MetricsReporterWithMockedChannels,
    SimpleMetricReporter,
    verify_models_equivalent_after_training,
)
from flsim.tests.utils import SampleNetHive
from flsim.trainers.async_trainer import AsyncTrainer, AsyncTrainerConfig
from flsim.trainers.sync_trainer import SyncTrainer, SyncTrainerConfig
from flsim.utils.config_utils import fl_config_from_json
from flsim.utils.fl.common import FLModelParamUtils
from flsim.utils.sample_model import DummyAlphabetFLModel
from flsim.utils.tests.helpers.sync_trainer_test_utils import create_sync_trainer
from flsim.utils.tests.helpers.test_data_utils import DummyAlphabetDataset
from flsim.utils.tests.helpers.test_utils import FLTestUtils
from flsim.utils.timing.training_duration_distribution import (
    PerExampleGaussianDurationDistributionConfig,
)
from hydra.experimental import compose, initialize
from hydra.utils import instantiate
from omegaconf import OmegaConf

CONFIG_PATH = "test_resources"

SYNC_TRAINER_JSON = f"{CONFIG_PATH}/sync_trainer.json"
ASYNC_TRAINER_JSON = f"{CONFIG_PATH}/async_trainer.json"
SYNC_TRAINER_WITH_DP_JSON = f"{CONFIG_PATH}/sync_trainer_with_dp.json"
SYNC_TRAINER_WRONG_JSON = f"{CONFIG_PATH}/sync_trainer_wrong_dp_config.json"
SYNC_TRAINER_WITH_SECAGG_JSON = f"{CONFIG_PATH}/sync_trainer_with_secagg.json"

SYNC_TRAINER_YAML = "sync_trainer"
ASYNC_TRAINER_YAML = "async_trainer"
SYNC_TRAINER_WITH_DP_YAML = "sync_trainer_with_dp"
SYNC_TRAINER_WRONG_YAML = "sync_trainer_wrong_dp_config"
SYNC_TRAINER_WITH_SECAGG_YAML = "sync_trainer_with_secagg"


class TestTrainer:
    @pytest.mark.parametrize(
        "json_file_name,trainer_class ",
        [
            (SYNC_TRAINER_JSON, SyncTrainer),
            (ASYNC_TRAINER_JSON, AsyncTrainer),
            (SYNC_TRAINER_WITH_DP_JSON, SyncTrainer),
            (SYNC_TRAINER_WITH_SECAGG_JSON, SyncTrainer),
        ],
    )
    def test_trainer_creation_from_json_config(
        self, json_file_name: str, trainer_class: type
    ):
        trainer = None
        file_path = pkg_resources.resource_filename(__name__, json_file_name)
        with open(file_path, "r") as parameters_file:
            json_cfg = json.load(parameters_file)
        cfg = fl_config_from_json(json_cfg)
        trainer = instantiate(
            cfg.trainer,
            model=DummyAlphabetFLModel(),
            cuda_enabled=False,
        )
        assertIsInstance(trainer, trainer_class)

    def test_trainer_sync_server_creation_from_json_config(self):
        file_path = pkg_resources.resource_filename(__name__, SYNC_TRAINER_JSON)
        with open(file_path, "r") as parameters_file:
            json_cfg = json.load(parameters_file)
        cfg = fl_config_from_json(json_cfg)
        trainer = instantiate(
            cfg.trainer,
            model=DummyAlphabetFLModel(),
            cuda_enabled=False,
        )
        assertIsInstance(trainer.server._optimizer, torch.optim.Adam)

    @pytest.mark.parametrize(
        "yaml_file_name,trainer_class ",
        [
            (SYNC_TRAINER_YAML, SyncTrainer),
            (ASYNC_TRAINER_YAML, AsyncTrainer),
            (SYNC_TRAINER_WITH_DP_YAML, SyncTrainer),
            (SYNC_TRAINER_WITH_SECAGG_YAML, SyncTrainer),
        ],
    )
    def test_trainer_creation_from_yaml_config(
        self, yaml_file_name: str, trainer_class: type
    ):
        trainer = None
        with initialize(config_path=CONFIG_PATH):
            cfg = compose(config_name=yaml_file_name)
            trainer = instantiate(
                cfg.trainer,
                model=DummyAlphabetFLModel(),
                cuda_enabled=False,
            )
        assertIsInstance(trainer, trainer_class)

    def test_async_trainer_with_dp_creation_from_json_config(self):
        trainer = None
        file_path = pkg_resources.resource_filename(__name__, ASYNC_TRAINER_JSON)
        with open(file_path, "r") as parameters_file:
            json_cfg = json.load(parameters_file)
        cfg = fl_config_from_json(json_cfg)
        trainer = instantiate(
            cfg.trainer,
            model=DummyAlphabetFLModel(),
            cuda_enabled=False,
        )
        assertIsInstance(trainer, AsyncTrainer)
        assertTrue(trainer.aggregator.is_private)

    def test_async_trainer_with_dp_creation_from_yaml_config(self):
        trainer = None
        with initialize(config_path=CONFIG_PATH):
            cfg = compose(config_name=ASYNC_TRAINER_YAML)
            trainer = instantiate(
                cfg.trainer,
                model=DummyAlphabetFLModel(),
                cuda_enabled=False,
            )
        assertIsInstance(trainer, AsyncTrainer)
        assertTrue(trainer.aggregator.is_private)

    def test_global_model_unchanged_after_metrics_reporting(self):
        """
        reporting metrics after aggregation should NOT update the global model
        """
        # mock 26 rows
        shard_size = 4
        local_batch_size = 4
        world_size = 1
        dummy_dataset = DummyAlphabetDataset()
        fl_data_sharder = SequentialSharder(examples_per_shard=shard_size)
        data_loader = FLDatasetDataLoaderWithBatch(
            dummy_dataset,
            dummy_dataset,
            dummy_dataset,
            fl_data_sharder,
            local_batch_size,
            local_batch_size,
            local_batch_size,
        )

        global_model = DummyAlphabetFLModel()

        local_optimizer_lr = 0.1
        metrics_reporter = FakeMetricReporter()
        users_per_round = 2
        sync_trainer_no_report = create_sync_trainer(
            model=global_model,
            local_lr=local_optimizer_lr,
            users_per_round=users_per_round,
            epochs=3,
            user_epochs_per_round=2,
        )

        sync_trainer_report = create_sync_trainer(
            model=copy.deepcopy(global_model),
            local_lr=local_optimizer_lr,
            users_per_round=users_per_round,
            epochs=3,
            user_epochs_per_round=2,
            report_train_metrics=True,
        )
        sync_trainer_report.cfg.report_train_metrics_after_aggregation = True

        data_provider = FLDataProviderFromList(
            data_loader.fl_train_set(),
            data_loader.fl_eval_set(),
            data_loader.fl_test_set(),
            global_model,
        )

        # training with reporting the train metrics after aggregation
        modules = []
        for trainer in [sync_trainer_report, sync_trainer_no_report]:
            model, _ = trainer.train(
                data_provider,
                metrics_reporter,
                num_total_users=data_provider.num_users(),
                distributed_world_size=world_size,
            )
            modules.append(model.fl_get_module())

        # make sure metrics reporting after aggregation does not change global model
        assertEqual(FLModelParamUtils.get_mismatched_param(modules), "")

    def test_client_optimizer_creation_from_config(self):
        """
        Test if trainer can instanciate the correct client optimizer from config
        """
        for optimizer_config, trainer_type in zip(
            [LocalOptimizerSGDConfig(lr=1.0), LocalOptimizerFedProxConfig(mu=1.0)],
            [SyncTrainer, AsyncTrainer],
        ):
            config = (
                SyncTrainerConfig(client=ClientConfig(optimizer=optimizer_config))
                if trainer_type == SyncTrainer
                else AsyncTrainerConfig(client=ClientConfig(optimizer=optimizer_config))
            )
            trainer = instantiate(
                config, model=DummyAlphabetFLModel(), cuda_enabled=False
            )
            assertTrue(isinstance(trainer, trainer_type))

    @pytest.mark.parametrize(
        "config,trainer_type ",
        [
            (
                SyncServerConfig(
                    server_optimizer=FedAvgWithLROptimizerConfig(lr=1.0, momentum=0.0)
                ),
                SyncTrainer,
            ),
            (FedAdamAsyncAggregatorConfig(beta1=0.1), AsyncTrainer),
        ],
    )
    def test_server_optimizer_creation_from_config(self, config, trainer_type):
        """
        Test if trainer can instanciate correct aggregator config
        """
        config = (
            SyncTrainerConfig(server=config)
            if trainer_type == SyncTrainer
            else AsyncTrainerConfig(aggregator=config)
        )
        trainer = instantiate(config, model=DummyAlphabetFLModel(), cuda_enabled=False)
        assertTrue(isinstance(trainer, trainer_type))

    def test_same_training_results_with_post_aggregation_reporting(self):
        """
        creat two training instances,
        one with report_train_metrics_after_aggregation=True, another with False,
        check training results match
        """
        torch.manual_seed(1)
        # mock 26 rows
        shard_size = 4
        local_batch_size = 4
        world_size = 1
        dummy_dataset = DummyAlphabetDataset()
        fl_data_sharder = SequentialSharder(examples_per_shard=shard_size)
        data_loader = FLDatasetDataLoaderWithBatch(
            dummy_dataset,
            dummy_dataset,
            dummy_dataset,
            fl_data_sharder,
            local_batch_size,
            local_batch_size,
            local_batch_size,
        )

        users_per_round = 2
        local_optimizer_lr = 0.1
        torch.manual_seed(1)
        # first training instance
        global_model_1 = DummyAlphabetFLModel()
        metrics_reporter_1 = FakeMetricReporter()
        sync_trainer_1 = create_sync_trainer(
            model=global_model_1,
            local_lr=local_optimizer_lr,
            users_per_round=users_per_round,
            epochs=3,
            user_epochs_per_round=2,
        )
        data_provider = FLDataProviderFromList(
            data_loader.fl_train_set(),
            data_loader.fl_eval_set(),
            data_loader.fl_test_set(),
            global_model_1,
        )
        assertEqual(data_provider.num_users(), data_loader.num_total_users)
        # training with reporting the train metrics after aggregation
        sync_trainer_1.cfg.report_train_metrics = True
        sync_trainer_1.cfg.report_train_metrics_after_aggregation = True
        global_model_1, best_metric_1 = sync_trainer_1.train(
            data_provider,
            metrics_reporter_1,
            num_total_users=data_provider.num_users(),
            distributed_world_size=world_size,
        )

        torch.manual_seed(1)
        # second training instance
        global_model_2 = DummyAlphabetFLModel()
        metrics_reporter_2 = FakeMetricReporter()
        sync_trainer_2 = create_sync_trainer(
            model=global_model_2,
            users_per_round=users_per_round,
            local_lr=local_optimizer_lr,
            epochs=3,
            user_epochs_per_round=2,
        )
        # training without reporting the train metrics after aggregation
        sync_trainer_2.cfg.report_train_metrics = True
        sync_trainer_2.cfg.report_train_metrics_after_aggregation = False
        global_model_2, best_metric_2 = sync_trainer_2.train(
            data_provider,
            metrics_reporter_2,
            num_total_users=data_provider.num_users(),
            distributed_world_size=world_size,
        )

        # check two training instance produce same results
        assertEqual(
            FLModelParamUtils.get_mismatched_param(
                [global_model_1.fl_get_module(), global_model_2.fl_get_module()]
            ),
            "",
        )

    def test_different_metrics_with_aggregation_client_reporting(self):
        """
        creat two training instances,
        one with use_train_clients_for_aggregation_metrics=True, another with False,
        check that the aggregation eval metrics are different
        """
        torch.manual_seed(1)
        # mock 26 rows
        shard_size = 4
        local_batch_size = 4
        world_size = 1
        dummy_dataset = DummyAlphabetDataset()
        fl_data_sharder = SequentialSharder(examples_per_shard=shard_size)
        data_loader = FLDatasetDataLoaderWithBatch(
            dummy_dataset,
            dummy_dataset,
            dummy_dataset,
            fl_data_sharder,
            local_batch_size,
            local_batch_size,
            local_batch_size,
        )

        users_per_round = 2
        local_optimizer_lr = 0.1
        torch.manual_seed(1)
        # first training instance
        global_model_1 = DummyAlphabetFLModel()
        metrics_reporter_1 = SimpleMetricReporter()
        sync_trainer_1 = create_sync_trainer(
            model=global_model_1,
            local_lr=local_optimizer_lr,
            users_per_round=users_per_round,
            epochs=3,
            user_epochs_per_round=2,
        )
        data_provider = FLDataProviderFromList(
            data_loader.fl_train_set(),
            data_loader.fl_eval_set(),
            data_loader.fl_test_set(),
            global_model_1,
        )
        assertEqual(data_provider.num_users(), data_loader.num_total_users)
        # training with using training clients for aggregation training metrics
        sync_trainer_1.cfg.report_train_metrics_after_aggregation = True
        sync_trainer_1.cfg.use_train_clients_for_aggregation_metrics = True
        sync_trainer_1.cfg.report_train_metrics = True
        global_model_1, best_metric_1 = sync_trainer_1.train(
            data_provider,
            metrics_reporter_1,
            num_total_users=data_provider.num_users(),
            distributed_world_size=world_size,
        )

        torch.manual_seed(1)
        # second training instance
        global_model_2 = DummyAlphabetFLModel()
        metrics_reporter_2 = SimpleMetricReporter()
        sync_trainer_2 = create_sync_trainer(
            model=global_model_2,
            users_per_round=users_per_round,
            local_lr=local_optimizer_lr,
            epochs=3,
            user_epochs_per_round=2,
        )
        # training with using training clients for aggregation training metrics
        sync_trainer_2.cfg.report_train_metrics_after_aggregation = True
        sync_trainer_2.cfg.use_train_clients_for_aggregation_metrics = False
        sync_trainer_2.cfg.report_train_metrics = True
        global_model_2, best_metric_2 = sync_trainer_2.train(
            data_provider,
            metrics_reporter_2,
            num_total_users=data_provider.num_users(),
            distributed_world_size=world_size,
        )

        # Check that the reported metrics are different
        for batch_metrics_1, batch_metrics_2 in zip(
            metrics_reporter_1.batch_metrics, metrics_reporter_2.batch_metrics
        ):
            if (
                batch_metrics_1.loss != batch_metrics_2.loss
                or batch_metrics_1.num_examples != batch_metrics_2.num_examples
            ):
                return
        assert True, "Batch metrics same whether using training or random clients"

    def test_one_user_sequential_user_equivalent(self):
        """
        test equivalence of the following scenario,

        1. one user who got all the example, and BS = all examples,
        local optimizer LR = 1.0
        2. number of users = number of examples, and one exampler per user,
        users per round = all users, use FedAvg
        """

        torch.manual_seed(1)
        # create dummy FL model on alphabet
        global_model = DummyAlphabetFLModel()
        global_model_init = copy.deepcopy(global_model)
        # will be used later to verify training indeed took place
        global_model_init_copy = copy.deepcopy(global_model)
        metrics_reporter = FakeMetricReporter()

        num_training_examples = 32
        # one user, who got 32 examples
        shard_size = num_training_examples
        local_batch_size = num_training_examples
        world_size = 1
        dummy_dataset = DummyAlphabetDataset(num_rows=num_training_examples)
        (
            data_provider,
            data_loader,
        ) = DummyAlphabetDataset.create_data_provider_and_loader(
            dummy_dataset, shard_size, local_batch_size, global_model
        )
        assertEqual(
            data_loader.num_total_users, math.ceil(num_training_examples / shard_size)
        )

        users_per_round = 1
        local_optimizer_lr = 1.0
        epochs = 5
        torch.manual_seed(1)
        sync_trainer = create_sync_trainer(
            model=global_model,
            local_lr=local_optimizer_lr,
            users_per_round=users_per_round,
            epochs=epochs,
        )
        one_user_global_model, _eval_metric_one_user = sync_trainer.train(
            data_provider,
            metrics_reporter,
            num_total_users=data_provider.num_users(),
            distributed_world_size=world_size,
        )
        metrics_reporter.reset()

        # 32 users, one example each
        shard_size = 1
        local_batch_size = 1
        world_size = 1
        dummy_dataset = DummyAlphabetDataset(num_rows=num_training_examples)
        (
            data_provider,
            data_loader,
        ) = DummyAlphabetDataset.create_data_provider_and_loader(
            dummy_dataset, shard_size, local_batch_size, global_model
        )
        assertEqual(
            data_loader.num_total_users, math.ceil(num_training_examples / shard_size)
        )
        assertEqual(data_provider.num_users(), data_loader.num_total_users)
        # select all users to train in each round
        users_per_round = data_provider.num_users()
        torch.manual_seed(1)
        sync_trainer = create_sync_trainer(
            model=global_model_init,
            local_lr=local_optimizer_lr,
            users_per_round=users_per_round,
            epochs=epochs,
        )
        all_users_global_model, _eval_metric_all_user = sync_trainer.train(
            data_provider,
            metrics_reporter,
            num_total_users=data_provider.num_users(),
            distributed_world_size=world_size,
        )
        assertEqual(
            verify_models_equivalent_after_training(
                one_user_global_model,
                all_users_global_model,
                global_model_init_copy,
                rel_epsilon=1e-4,
                abs_epsilon=1e-6,
            ),
            "",
        )

    def test_training_with_armijo_line_search(self):
        """
        test Armijo line-search for local LR scheduling

        using shrinking factor = 1.0  == constant LR
        """

        torch.manual_seed(1)
        # create dummy FL model on alphabet
        global_model = DummyAlphabetFLModel()
        global_model_init = copy.deepcopy(global_model)
        # will be used later to verify training indeed took place
        global_model_init_copy = copy.deepcopy(global_model)
        metrics_reporter = FakeMetricReporter()

        # one user, who got 26 examples
        shard_size = 26
        local_batch_size = 26
        world_size = 1
        dummy_dataset = DummyAlphabetDataset()
        (
            data_provider,
            data_loader,
        ) = DummyAlphabetDataset.create_data_provider_and_loader(
            dummy_dataset, shard_size, local_batch_size, global_model
        )
        assertEqual(data_loader.num_total_users, math.ceil(26 / shard_size))
        assertEqual(data_provider.num_users(), data_loader.num_total_users)
        users_per_round = 1
        local_optimizer_lr = 1.0
        epochs = 5
        torch.manual_seed(1)
        sync_trainer = create_sync_trainer(
            model=global_model,
            local_lr=local_optimizer_lr,
            users_per_round=users_per_round,
            epochs=epochs,
        )

        constant_lr_model, _ = sync_trainer.train(
            data_provider,
            metrics_reporter,
            num_total_users=data_provider.num_users(),
            distributed_world_size=world_size,
        )
        metrics_reporter.reset()

        torch.manual_seed(1)
        # training with Armijo line-search LR scheduler with
        # shrinking factor = 1.0 (no shrinking)
        local_lr_scheduler_config = ArmijoLineSearchSchedulerConfig(
            shrinking_factor=1.0
        )
        sync_trainer_with_scheduler = create_sync_trainer(
            model=global_model_init,
            local_lr=local_optimizer_lr,
            users_per_round=users_per_round,
            epochs=epochs,
            local_lr_scheduler=local_lr_scheduler_config,
        )

        armijo_ls_model, _ = sync_trainer_with_scheduler.train(
            data_provider,
            metrics_reporter,
            num_total_users=data_provider.num_users(),
            distributed_world_size=world_size,
        )
        metrics_reporter.reset()
        assertEqual(
            verify_models_equivalent_after_training(
                constant_lr_model,
                armijo_ls_model,
                global_model_init_copy,
                rel_epsilon=1e-6,
                abs_epsilon=1e-6,
            ),
            "",
        )

    def _test_fl_nonfl_equivalent(
        self,
        num_examples: int,
        num_fl_users: int,
        epochs: int,
        local_lr: float,
        server_config: SyncServerConfig,
    ):
        """
        Given:
            data_for_fl={user1:batch1, user2:batch2}
            data_for_non_fl={batch1, batch2}

        Check that the following produce the same trained model:
        1. FL training, 1 user per round. global_opt=SGD, local_lr=x, global_lr=x
        2. Non-FL training, opt=SGD, lr=x
        """
        torch.manual_seed(1)
        # create dummy FL model on alphabet
        global_model = DummyAlphabetFLModel()
        # will be used later to verify training indeed took place
        global_model_init_copy = copy.deepcopy(global_model)
        # num_fl_users users, each with num_examples/num_fl_users training examples
        assertTrue(
            num_examples % num_fl_users == 0,
            f"Expect num_examples({num_examples}) to be multiple of num_fl_users({num_fl_users})",
        )
        shard_size = num_examples // num_fl_users
        batch_size = num_examples // num_fl_users
        world_size = 1
        dummy_dataset = DummyAlphabetDataset(num_examples)
        (
            data_provider,
            data_loader,
        ) = DummyAlphabetDataset.create_data_provider_and_loader(
            dummy_dataset, shard_size, batch_size, global_model
        )
        assertEqual(data_loader.num_total_users, math.ceil(num_examples / shard_size))
        assertEqual(data_provider.num_users(), data_loader.num_total_users)

        torch.manual_seed(1)

        sync_trainer = create_sync_trainer(
            model=global_model,
            local_lr=local_lr,
            users_per_round=1,
            epochs=epochs,
            do_eval=False,
            server_config=server_config,
        )
        fl_model, _ = sync_trainer.train(
            data_provider,
            metric_reporter=FakeMetricReporter(),
            num_total_users=data_provider.num_users(),
            distributed_world_size=world_size,
        )

        # non-FL training
        dummy_dataset = DummyAlphabetDataset(num_examples)
        data_loader = torch.utils.data.DataLoader(
            dummy_dataset, batch_size=batch_size, shuffle=False
        )
        nonfl_model = copy.deepcopy(global_model_init_copy)

        optimizer = instantiate(
            config=server_config.server_optimizer,
            model=nonfl_model.fl_get_module(),
        )

        FLTestUtils.run_nonfl_training(
            model=nonfl_model,
            optimizer=optimizer,
            data_loader=data_loader,
            epochs=epochs,
        )

        error_msg = verify_models_equivalent_after_training(
            fl_model,
            nonfl_model,
            global_model_init_copy,
            rel_epsilon=1e-4,
            abs_epsilon=1e-6,
        )
        assertEmpty(error_msg, msg=error_msg)

    def test_fl_nonfl_equivalent_global_optimizer_sgd(self):
        """
        Given:
            batch_size=13
            data_for_fl={user1:batch1, user2:batch2}
            data_for_non_fl={batch1, batch2}

        Check that the following produce the same trained model:
        1. FL training, 1 user per round. global_opt=SGD, local_lr=x, global_lr=x
        2. Non-FL training, opt=SGD, lr=x
        """
        self._test_fl_nonfl_equivalent(
            num_examples=26,
            num_fl_users=2,
            epochs=5,
            local_lr=1.0,
            server_config=SyncServerConfig(
                server_optimizer=FedAvgWithLROptimizerConfig(lr=1.0, momentum=0.0),
                active_user_selector=SequentialActiveUserSelectorConfig(),
            ),
        )

    def test_fl_nonfl_equivalent_global_optimizer_adam(self):
        """
        Given:
            batch_size=16 (bs=num_examples/num_fl_users)
            data_for_fl={user1:batch1, user2:batch2}
            data_for_non_fl={batch1, batch2}

        Check that the following produce the same trained model:
        1. FL training, 1 user per round. GlobalOpt=Adam, local_lr=1.0, global_lr=x
        2. Non-FL training, opt=Adam, lr=x
        """
        self._test_fl_nonfl_equivalent(
            num_examples=32,
            num_fl_users=2,
            epochs=5,
            local_lr=1.0,
            server_config=SyncServerConfig(
                server_optimizer=FedAdamOptimizerConfig(lr=0.001, eps=1e-2),
                active_user_selector=SequentialActiveUserSelectorConfig(),
            ),
        )

    def test_client_overselection(self):
        """
        test client overselection by equivalence of the two setups:

        1. two users in the population, user 0 has two more examples than user 1.
        UPR = 1, dropout_rate = 0.5. Hence (1/0.5=2) users will be over-selected.
        simulate training time with mean = 1.0, std = 0.0 per example.
        Hence user 0 will always be dropped due. Trains with 1 FL epochs. rounds
        per epoch = (epoch / UPR = 2), hence user 1 will be trained twice.

        2. single user in the population, which is user 1 the the setting above.
        UPR = 1, dropout = 1.0. two FL epochs.
        """
        torch.manual_seed(1)
        # create dummy FL model on alphabet
        global_model = DummyAlphabetFLModel()
        # keep a copy of initial model to trigger another training instance
        global_model_init = copy.deepcopy(global_model)
        # dummy alphabet dataset
        dummy_dataset = DummyAlphabetDataset()
        # two users, one gets 1 more example than the other
        shard_size = len(dummy_dataset) // 2 + 1
        local_batch_size = 2
        (
            data_provider,
            data_loader,
        ) = DummyAlphabetDataset.create_data_provider_and_loader(
            dummy_dataset, shard_size, local_batch_size, global_model
        )
        assertEqual(data_provider.num_users(), data_loader.num_total_users)
        # assert first user gets (dummy_dataset.num_rows / 2) + 1 data point,
        # the second user gets (dummy_dataset.num_rows / 2) - 1 data point
        assertEqual(data_provider[0].num_examples(), dummy_dataset.num_rows / 2 + 1)
        assertEqual(data_provider[1].num_examples(), dummy_dataset.num_rows / 2 - 1)
        # shared trainer config between two training instance
        users_per_round = 1
        local_optimizer_lr = 1.0
        epochs = 1
        torch.manual_seed(1)
        # training with overselection
        dropout_rate = 0.5
        sync_trainer_overselection = create_sync_trainer(
            model=global_model,
            local_lr=local_optimizer_lr,
            users_per_round=users_per_round,
            epochs=epochs,
            dropout_rate=dropout_rate,
            timeout_simulator_config=GaussianTimeOutSimulatorConfig(
                duration_distribution_generator=PerExampleGaussianDurationDistributionConfig(
                    training_duration_mean=1.0, training_duration_sd=0.0
                ),
                timeout_wall_per_round=999999,
                fl_stopping_time=9999999,
            ),
        )
        model_with_overselection, _ = sync_trainer_overselection.train(
            data_provider,
            metric_reporter=FakeMetricReporter(),
            num_total_users=data_provider.num_users(),
            distributed_world_size=1,
        )
        # another training instance: only user[1] in the user population
        # removing user 0 from dataset, assign user 1 to be user 0
        data_provider._users[0] = copy.deepcopy(data_provider._users[1])
        data_provider._users.pop(1)
        # only a single user after remove user 0
        assertEqual(data_provider.num_users(), 1)
        global_model = copy.deepcopy(global_model_init)
        torch.manual_seed(1)
        dropout_rate = 1.0
        epochs = 2
        sync_trainer_single_user = create_sync_trainer(
            model=global_model,
            local_lr=local_optimizer_lr,
            users_per_round=users_per_round,
            epochs=epochs,
            dropout_rate=dropout_rate,
        )
        model_single_user, _ = sync_trainer_single_user.train(
            data_provider,
            metric_reporter=FakeMetricReporter(),
            num_total_users=data_provider.num_users(),
            distributed_world_size=1,
        )
        assertEqual(
            verify_models_equivalent_after_training(
                model_with_overselection,
                model_single_user,
                global_model_init,
                rel_epsilon=1e-4,
                abs_epsilon=1e-6,
            ),
            "",
        )

    def test_partial_update_from_clients(self):
        """
        test the equivalence of these two training instance
        1. UPR=1. User dataset has 52 characters [a,b,c,d .... x, y ,z, a, b, c... x,y,z],
        where each character appears twice in order. Timeout limit is set to
        just enough for training half of the dataset

        2. UPR=1. User daraset has 26 characters [a,b,c,d ...,x,y,z]. no timeout
        """
        torch.manual_seed(1)
        # create dummy FL model on alphabet
        global_model = DummyAlphabetFLModel()
        # keep a copy of initial model to trigger another training instance
        global_model_init = copy.deepcopy(global_model)
        # dummy alphabet dataset, getting each character twrice
        num_rows = 52
        dummy_dataset = DummyAlphabetDataset(num_rows)
        # a single user getting all examples
        shard_size = len(dummy_dataset)
        local_batch_size = 2
        (
            data_provider,
            data_loader,
        ) = DummyAlphabetDataset.create_data_provider_and_loader(
            dummy_dataset, shard_size, local_batch_size, global_model
        )
        # shared trainer config between two training instance
        users_per_round = 1
        local_optimizer_lr = 1.0
        epochs = 1
        torch.manual_seed(1)
        # training time just enought for 26 examples, although user has
        # 52 examples
        sync_trainer_timeout = create_sync_trainer(
            model=copy.deepcopy(global_model),
            local_lr=local_optimizer_lr,
            users_per_round=users_per_round,
            epochs=epochs,
            timeout_simulator_config=GaussianTimeOutSimulatorConfig(
                duration_distribution_generator=PerExampleGaussianDurationDistributionConfig(
                    training_duration_mean=1.0, training_duration_sd=0.0
                ),
                timeout_wall_per_round=26,
                fl_stopping_time=9999999,
            ),
        )
        model_with_timeout, _ = sync_trainer_timeout.train(
            data_provider,
            metric_reporter=FakeMetricReporter(),
            num_total_users=data_provider.num_users(),
            distributed_world_size=1,
        )
        # dummy alphabet dataset, getting each character once
        num_rows = 26
        torch.manual_seed(1)
        dummy_dataset = DummyAlphabetDataset(num_rows)
        # a single user getting all examples
        shard_size = len(dummy_dataset)
        local_batch_size = 2
        (
            data_provider,
            data_loader,
        ) = DummyAlphabetDataset.create_data_provider_and_loader(
            dummy_dataset, shard_size, local_batch_size, global_model
        )
        torch.manual_seed(1)
        # training time just enought for 26 examples
        sync_trainer_timeout = create_sync_trainer(
            model=copy.deepcopy(global_model),
            local_lr=local_optimizer_lr,
            users_per_round=users_per_round,
            epochs=epochs,
        )
        model_no_timeout, _ = sync_trainer_timeout.train(
            data_provider,
            metric_reporter=FakeMetricReporter(),
            num_total_users=data_provider.num_users(),
            distributed_world_size=1,
        )
        assertEqual(
            verify_models_equivalent_after_training(
                model_with_timeout,
                model_no_timeout,
                global_model_init,
                rel_epsilon=1e-4,
                abs_epsilon=1e-6,
            ),
            "",
        )

    def test_client_metric_reporting(self):
        """
        Test that per-client reporting reports exactly every
        ``client_metrics_reported_per_epoch`` as defined in the config.
        """
        torch.manual_seed(1)
        # create dummy FL model on alphabet
        global_model = DummyAlphabetFLModel()
        # dummy alphabet dataset, getting each character twrice
        num_rows = 52
        num_users = 4
        dummy_dataset = DummyAlphabetDataset(num_rows)
        # a single user getting all examples
        shard_size = int(len(dummy_dataset) / num_users)
        local_batch_size = 2
        (
            data_provider,
            data_loader,
        ) = DummyAlphabetDataset.create_data_provider_and_loader(
            dummy_dataset, shard_size, local_batch_size, global_model
        )
        users_per_round = 1
        local_optimizer_lr = 1.0
        momentum = 0.9
        epochs = 6

        sync_trainer_with_client_reports = SyncTrainer(
            model=global_model,
            cuda_enabled=False,
            **OmegaConf.structured(
                SyncTrainerConfig(
                    epochs=epochs,
                    do_eval=True,
                    users_per_round=users_per_round,
                    always_keep_trained_model=False,
                    train_metrics_reported_per_epoch=1,
                    eval_epoch_frequency=1,
                    report_train_metrics=False,
                    report_train_metrics_after_aggregation=True,
                    client=ClientConfig(
                        epochs=1,
                        optimizer=LocalOptimizerSGDConfig(
                            lr=local_optimizer_lr, momentum=momentum
                        ),
                        max_clip_norm_normalized=None,
                    ),
                    server=SyncServerConfig(
                        server_optimizer=FedAvgWithLROptimizerConfig(
                            lr=1.0, momentum=0.0
                        ),
                        active_user_selector=SequentialActiveUserSelectorConfig(),
                    ),
                    report_client_metrics=True,
                    report_client_metrics_after_epoch=True,
                    client_metrics_reported_per_epoch=3,
                )
            ),
        )
        metrics_reporter = MetricsReporterWithMockedChannels()

        model, _ = sync_trainer_with_client_reports.train(
            data_provider,
            metric_reporter=metrics_reporter,
            num_total_users=data_provider.num_users(),
            distributed_world_size=1,
        )

        def count_word(result, word):
            return str(result).count(word)

        # If client_metrics_reported_per_epoch = 3 and number of epochs = 6
        # per client eval metrics should be reported twice.
        assertEqual(
            count_word(metrics_reporter.stdout_results, "Per_Client_Eval"),
            2,
            metrics_reporter.stdout_results,
        )

    def _get_tensorboard_results_from_training(
        self, num_total_users, num_epochs, users_per_round
    ) -> List[MockRecord]:
        # dataset has 26 rows
        assertTrue(num_total_users <= 26, "Can't have more than 26 users")
        torch.manual_seed(1)
        # 26 rows in data
        shard_size = int(math.ceil(26 / num_total_users))
        local_batch_size = 4
        dummy_dataset = DummyAlphabetDataset()
        fl_data_sharder = SequentialSharder(examples_per_shard=shard_size)
        data_loader = FLDatasetDataLoaderWithBatch(
            dummy_dataset,
            dummy_dataset,
            dummy_dataset,
            fl_data_sharder,
            local_batch_size,
            local_batch_size,
            local_batch_size,
        )

        torch.manual_seed(1)
        # first training instance
        global_model = DummyAlphabetFLModel()
        metrics_reporter = MetricsReporterWithMockedChannels()
        sync_trainer = create_sync_trainer(
            model=global_model,
            local_lr=0.1,
            users_per_round=users_per_round,
            epochs=num_epochs,
            user_epochs_per_round=1,
        )
        data_provider = FLDataProviderFromList(
            data_loader.fl_train_set(),
            data_loader.fl_eval_set(),
            data_loader.fl_test_set(),
            global_model,
        )
        assertEqual(data_provider.num_users(), data_loader.num_total_users)
        sync_trainer.cfg.report_train_metrics = True
        sync_trainer.cfg.report_train_metrics_after_aggregation = True
        global_model, best_metric = sync_trainer.train(
            data_provider,
            metrics_reporter,
            num_total_users=num_total_users,
            distributed_world_size=1,
        )
        return metrics_reporter.tensorboard_results

    def test_tensorboard_metrics_reporting_simple(self):
        """Train with tensorboard metrics reporter for one epoch, 5 rounds per epoch.
        Test for 2 things:
        a) Train, Aggregation and Eval metrics are reported once
        b) Metric reporting happens at the end of the epoch (when global_round=5)
        """
        num_total_users = 5
        num_epochs = 1
        users_per_round = 1
        tensorboard_results: List[
            MockRecord
        ] = self._get_tensorboard_results_from_training(
            num_total_users=num_total_users,
            num_epochs=num_epochs,
            users_per_round=users_per_round,
        )

        # ensure that train, aggregation and eval metrics are all reported once
        for stage in [
            TrainingStage.TRAINING,
            TrainingStage.AGGREGATION,
            TrainingStage.EVAL,
        ]:
            tag = f"Loss/{TrainingStage(stage).name.title()}"
            num_entries = sum(record.tag == tag for record in tensorboard_results)
            # we report once per epoch
            assertEqual(num_entries, 1)

        # training runs for 1 epoch, or 5 rounds
        # when train/eval results are reported, global_round_num should be 5
        global_steps_reported_expected = [5, 5, 5]  # noqa: F841
        global_steps_reported_actual = [
            record.global_step for record in tensorboard_results
        ]
        assertEqual(
            global_steps_reported_actual,
            global_steps_reported_expected,
            f"Actual global steps: {global_steps_reported_actual}, Expected global steps:{global_steps_reported_expected}",
        )

    def test_tensorboard_metrics_reporting_complex(self):
        """Train with tensorboard metrics reporter. Ensure eval and train metrics are
        correctly reported to tensorboard
        """
        # TODO: run with different values of num_epochs, num_total_users, users_per_round,
        # train_metrics_reported_per_epoch, eval_epoch_frequency
        num_total_users = 10
        num_epochs = 3
        users_per_round = 2
        tensorboard_results: List[
            MockRecord
        ] = self._get_tensorboard_results_from_training(
            num_total_users=num_total_users,
            num_epochs=num_epochs,
            users_per_round=users_per_round,
        )

        # ensure that train, aggregation and eval metrics are all reported once per epoch
        for stage in [
            TrainingStage.TRAINING,
            TrainingStage.AGGREGATION,
            TrainingStage.EVAL,
        ]:
            tag = f"Loss/{TrainingStage(stage).name.title()}"
            num_entries = sum(record.tag == tag for record in tensorboard_results)
            # we report once per epoch
            assertEqual(num_entries, num_epochs)

        rounds_per_epoch = int(math.ceil(num_total_users / users_per_round))
        # example. if num_epochs=3, and rounds_per_epoch=5, global_steps_at_epoch_end will be [5, 10, 15]
        epochs = list(range(1, num_epochs + 1))
        global_steps_at_epoch_end = [e * rounds_per_epoch for e in epochs]

        # we report 3 metrics in each epoch: Training, Aggregation and Eval
        num_records_in_report = 3
        # example. if global_steps_at_epoch_end = [5, 10, 15], and num_records_in_report = 2,
        # then global_steps_reported_expected = [5, 5, 10,10, 15,15]
        # noqa: F841  variable not used, fixed in next diff
        global_steps_reported_expected = [
            step
            for step in global_steps_at_epoch_end
            for k in range(num_records_in_report)
        ]
        # noqa: F841  variable not used, fixed in next diff
        global_steps_reported_actual = [
            record.global_step for record in tensorboard_results
        ]
        assertEqual(
            global_steps_reported_actual,
            global_steps_reported_expected,
            f"Actual global steps: {global_steps_reported_actual}, Expected global steps:{global_steps_reported_expected}",
        )

    @pytest.mark.parametrize(
        "page_turn_freq,users_per_round, pages_used",
        [
            (0.99, 5, 6),
            (0.99, 10, 6),
            (0.50, 10, 9),
            (0.50, 5, 9),
        ],
    )
    def test_sync_trainer_with_page_data_provider(
        self, page_turn_freq, users_per_round, pages_used
    ):
        """
        Test if sync trainer works properly with paged data loader
        Assumptions:
            users_per_page = 10 total number of users = 50
            batch_size = 1
            train for 2 global epochs

        Cases:
        1. Base case: user_per_round is half of page_size, turn page after all users are used
        2. Boundary case: users_per_round = page_size, turn page after all users are used
        3. Boundary case: users_per_round = page_size, turn page after half of users in the page are used
        """
        id_col = "label"
        num_total_users = 50
        inline_config = InlineDatasetConfig(
            num_users=num_total_users,
            data_cols=["user_n"],
            id_col=id_col,
        )
        dataloader = create_dataloader(
            num_users_per_page=10,
            num_total_users=num_total_users,
            batch_size=1,
            inline_config=inline_config,
            sharding_col_name=id_col,
            use_nid=True,
        )
        global_model = SampleNetHive()
        data_provider = PagedDataProvider(
            dataloader, model=global_model, page_turn_freq=page_turn_freq
        )
        sync_trainer = create_sync_trainer(
            model=global_model,
            local_lr=0.1,
            users_per_round=users_per_round,
            epochs=1,
        )

        sync_trainer.train(
            data_provider,
            metric_reporter=FakeMetricReporter(),
            # Note: We're using num_total_users and
            # not data_provider.num_users()
            num_total_users=num_total_users,
            distributed_world_size=1,
        )
        assertEqual(data_provider.pages_used, pages_used)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

import copy
import itertools
import math
from tempfile import mkstemp
from unittest.mock import MagicMock

import numpy as np
import pytest
import torch
import torch.distributed as dist
import torch.multiprocessing as mp
from flsim.channels.base_channel import IdentityChannel
from flsim.channels.half_precision_channel import HalfPrecisionChannel
from flsim.channels.message import Message
from flsim.common.pytest_helper import assertEmpty, assertEqual
from flsim.privacy.common import PrivacySetting
from flsim.privacy.privacy_engine import GaussianPrivacyEngine
from flsim.servers.aggregator import AggregationType
from flsim.servers.sync_dp_servers import SyncDPSGDServerConfig
from flsim.servers.sync_servers import FedAvgOptimizerConfig, SyncServerConfig
from flsim.tests.utils import (
    create_model_with_value,
    model_parameters_equal_to_value,
    verify_models_equivalent_after_training,
    SampleNet,
)
from flsim.utils.distributed.fl_distributed import FLDistributedUtils
from hydra.utils import instantiate


def init_process(
    rank,
    world_size,
    server,
    models,
    file_loc,
    pipe,
):
    FLDistributedUtils.dist_init(
        rank=rank,
        world_size=world_size,
        init_method=f"file://{file_loc}",
        use_cuda=False,
    )
    server.init_round()
    for i, m in enumerate(models):
        if i % world_size == rank:
            weight = i + 1
            server.receive_update_from_client(
                Message(model=SampleNet(m), weight=weight)
            )

    server.step()
    sums, weights = 0.0, 0.0
    all_sum = [
        (p.sum(), p.numel()) for p in server.global_model.fl_get_module().parameters()
    ]
    for s, w in all_sum:
        sums += float(s)
        weights += float(w)
    pipe.send(sums / weights)
    dist.destroy_process_group()


def run_multiprocess_server_test(server, num_processes=1, num_models=4):
    _, tmpfile = mkstemp(dir="/tmp")
    pipe_out, pipe_in = mp.Pipe(False)
    models = [create_model_with_value(1.0) for i in range(num_models)]

    processes = []
    results = []
    FLDistributedUtils.WORLD_SIZE = num_processes
    for pid in range(num_processes):
        p = mp.Process(
            target=init_process,
            args=(
                pid,
                num_processes,
                server,
                models,
                tmpfile,
                pipe_in,
            ),
        )
        p.start()
        processes.append(p)
        results.append(pipe_out)
    for p in processes:
        p.join()
    res = [r.recv() for r in results]
    return res


class TestSyncDPSGDServer:
    def _get_num_params(self, model):
        return sum(p.numel() for p in model.parameters())

    def _create_server(
        self,
        server_model,
        num_rounds,
        num_clients,
        clipping_value,
        noise_multiplier,
        channel=None,
    ):
        server = instantiate(
            SyncDPSGDServerConfig(
                aggregation_type=AggregationType.AVERAGE,
                server_optimizer=FedAvgOptimizerConfig(),
                privacy_setting=PrivacySetting(
                    clipping_value=clipping_value,
                    noise_multiplier=noise_multiplier,
                    noise_seed=0,
                ),
            ),
            global_model=server_model,
            channel=channel,
        )
        server.select_clients_for_training(
            num_total_users=num_rounds * num_clients, users_per_round=num_clients
        )
        return server

    @pytest.mark.parametrize(
        "clipping_value, num_clients", itertools.product([1, 1e10], [1, 10])
    )
    def test_no_noise_with_clip(self, clipping_value, num_clients) -> None:
        """
        Test DP-SGD with no noise and with user norm clipping.
        """
        server_model = SampleNet(create_model_with_value(0))
        num_rounds = 10
        delta_param = 1.0
        num_params = self._get_num_params(server_model.fl_get_module())

        clipped_delta = math.sqrt(clipping_value ** 2 / num_params)
        expected_value = -float(min(np.average(clipped_delta), delta_param))
        server = self._create_server(
            server_model, num_rounds, num_clients, clipping_value, noise_multiplier=0
        )

        for round_num in range(num_rounds):
            server.init_round()
            for _ in range(num_clients):
                delta = SampleNet(create_model_with_value(delta_param))
                server.receive_update_from_client(Message(model=delta, weight=1.0))
            server.step()
            error_msg = model_parameters_equal_to_value(
                server_model, expected_value * (round_num + 1)
            )
            assertEmpty(error_msg, msg=error_msg)

    @pytest.mark.parametrize(
        "clipping_value, noise, num_clients",
        itertools.product([1], [1], [1, 10]),
    )
    def test_noise_and_clip(self, clipping_value, noise, num_clients) -> None:
        """
        Test user-level DP-SGD.
        We assume the following:
            1. Server model init  at 0
            2. Trains for 10 rounds and take the simple average of the client updates.
            3. The learning rate = 1.0.
            4. The norm of each user delta is greater than the clipping value

        The DP-SGD update rule is: w_t = w_t-1 - lr * (avg(grad) + sgd_noise)
        With the above assumptions, w_t = 0 - 1.0 * (avg(grad) + sgd_noise) = -(avg(grad) + sgd_noise)
        """
        server_model = SampleNet(create_model_with_value(0))
        num_rounds = 10

        num_params = self._get_num_params(server_model.fl_get_module())
        clipped_delta = math.sqrt(clipping_value ** 2 / num_params)

        server = self._create_server(
            server_model,
            num_rounds,
            num_clients,
            clipping_value,
            noise_multiplier=noise,
        )

        GaussianPrivacyEngine._generate_noise = MagicMock(
            side_effect=lambda size, sensitivity: torch.ones(size) * sensitivity
        )
        for round_num in range(num_rounds):
            server.init_round()
            for _ in range(num_clients):
                delta = SampleNet(create_model_with_value(1.0))
                server.receive_update_from_client(Message(model=delta, weight=1.0))
            server.step()

            expected_value = float(
                -np.average(clipped_delta) - (noise * clipping_value / num_clients)
            )
            error_msg = model_parameters_equal_to_value(
                server_model, expected_value * (round_num + 1)
            )
            assertEmpty(error_msg, msg=error_msg)

    def test_no_noise_no_clip(self):
        """
        Test that DP-SGD server with no clipping and no noise is the same as vanilla SyncServer
        """
        global_value = 0
        client_value = 1.0
        dp_model = SampleNet(create_model_with_value(global_value))
        no_dp_model = SampleNet(create_model_with_value(global_value))
        num_rounds = 10
        num_clients = 10

        dp_server = self._create_server(
            dp_model,
            num_rounds,
            num_clients=num_clients,
            clipping_value=1e10,
            noise_multiplier=0,
        )
        no_dp_server = instantiate(
            SyncServerConfig(
                aggregation_type=AggregationType.AVERAGE,
                server_optimizer=FedAvgOptimizerConfig(),
            ),
            global_model=no_dp_model,
        )

        for _ in range(num_rounds):
            no_dp_server.init_round()
            dp_server.init_round()
            for _ in range(num_clients):
                dp_server.receive_update_from_client(
                    Message(
                        model=SampleNet(
                            create_model_with_value(global_value - client_value)
                        ),
                        weight=1.0,
                    )
                )
                no_dp_server.receive_update_from_client(
                    Message(
                        model=SampleNet(
                            create_model_with_value(global_value - client_value)
                        ),
                        weight=1.0,
                    )
                )
            dp_server.step()
            no_dp_server.step()

            error_msg = verify_models_equivalent_after_training(dp_model, no_dp_model)
            assertEmpty(error_msg, msg=error_msg)

    def test_noise_added_correctly(self):
        """
        Test where noise is a fixed value, 0.8
        update = global (all 0) - local (all 2.0) = all 2.0
        update norm = sqrt(num_params*delta^2)=sqrt(21*2^2)=sqrt(84)= 9.165
        and this will be clipped to clipping_value of 7, which
        means that the parameters of the clipped update will be all equal to sqrt(49/21)= 1.52
        w_t = w_t-1 - lr * (avg(grad) + sgd_noise)
        w_t = 0 - 1.0 * (avg(grad) + sgd_noise) = -(avg(grad) + sgd_noise) = -(1.52 + 0.8) = -2.32
        """
        num_clients = 10
        clipping_value = 7.0
        noise = 0.8
        global_value = 0.0
        client_value = -2.0

        server_model = SampleNet(create_model_with_value(global_value))

        num_params = self._get_num_params(server_model.fl_get_module())
        clipped_delta = math.sqrt(clipping_value ** 2 / num_params)
        expected_value = float(-np.average(clipped_delta) - noise)

        server = self._create_server(
            server_model,
            num_rounds=1,
            num_clients=num_clients,
            clipping_value=clipping_value,
            noise_multiplier=noise,
        )

        GaussianPrivacyEngine._generate_noise = MagicMock(return_value=noise)
        server.init_round()
        for _ in range(num_clients):
            delta = create_model_with_value(global_value - client_value)
            server.receive_update_from_client(Message(model=SampleNet(delta), weight=1))
        server.step()

        error_msg = model_parameters_equal_to_value(server_model, expected_value)
        assertEmpty(error_msg, msg=error_msg)

    @pytest.mark.parametrize(
        "channel",
        [HalfPrecisionChannel(), IdentityChannel()],
    )
    def test_dp_server_channel_integration(self, channel):
        """From Client to Server, the channel should quantize and then dequantize the message
        therefore there should be no change in the model
        """
        server = self._create_server(
            SampleNet(create_model_with_value(0)),
            num_rounds=1,
            num_clients=10,
            clipping_value=10,
            noise_multiplier=0,
            channel=channel,
        )
        delta = create_model_with_value(1)
        init = copy.deepcopy(delta)
        server.receive_update_from_client(Message(model=SampleNet(delta), weight=1.0))
        error_msg = verify_models_equivalent_after_training(delta, init)
        assertEmpty(error_msg, msg=error_msg)

    def test_sync_dp_server_with_multiple_processes(self):
        server = self._create_server(
            SampleNet(create_model_with_value(0)),
            num_rounds=1,
            num_clients=10,
            clipping_value=10,
            noise_multiplier=0,
        )
        results = run_multiprocess_server_test(server, num_processes=4, num_models=4)
        for r, v in zip(results, results[1:]):
            assertEqual(r, v)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

import abc
from dataclasses import dataclass
from typing import Optional, List

from flsim.active_user_selectors.simple_user_selector import (
    ActiveUserSelectorConfig,
    UniformlyRandomActiveUserSelectorConfig,
)
from flsim.channels.base_channel import (
    IdentityChannel,
    IFLChannel,
)
from flsim.channels.message import Message
from flsim.data.data_provider import IFLDataProvider
from flsim.interfaces.model import IFLModel
from flsim.optimizers.server_optimizers import (
    ServerOptimizerConfig,
    FedAvgOptimizerConfig,
    OptimizerType,
)
from flsim.servers.aggregator import AggregationType, Aggregator
from flsim.utils.config_utils import fullclassname, init_self_cfg
from flsim.utils.fl.common import FLModelParamUtils
from hydra.utils import instantiate
from omegaconf import OmegaConf


class ISyncServer(abc.ABC):
    """
    Interface for Sync servers, all sync server should
    implement this interface.
    Responsiblities:
        Wrapper for aggregator and optimizer.
        Collects client updates and sends to aggregator.
        Changes the global model using aggregator and optimizer.
    """

    @abc.abstractmethod
    def init_round(self):
        """
        Clears the buffer and zero out grad in optimizer
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def receive_update_from_client(self, message: Message):
        """
        Receives new update from client
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def step(self):
        """
        Update the global model
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def select_clients_for_training(
        self,
        num_total_users: int,
        users_per_round: int,
        data_provider: Optional[IFLDataProvider] = None,
        epoch: Optional[int] = None,
    ) -> List[int]:
        """
        Selects clients to participate in a round of training.

        The selection scheme depends on the underlying selector. This
        can include: random, sequential, high loss etc.

        Args:
            num_total_users ([int]): Number of total users (population size).
            users_per_round ([int]]): Number of users per round.
            data_provider (Optional[IFLDataProvider], optional): This is useful when the selection scheme
            is high loss. Defaults to None.
            epoch (Optional[int], optional): [description]. This is useful when the selection scheme
            is high loss. Defaults to None.

        Returns:
            List[int]: A list of client indicies
        """
        pass

    @property
    def global_model(self) -> IFLModel:
        """
        Returns the current global model
        """
        raise NotImplementedError()


class SyncServer(ISyncServer):
    def __init__(
        self,
        *,
        global_model: IFLModel,
        channel: Optional[IFLChannel] = None,
        **kwargs,
    ):
        init_self_cfg(
            self,
            component_class=__class__,  # pyre-fixme[10]: Name `__class__` is used but not defined.
            config_class=SyncServerConfig,
            **kwargs,
        )
        self._optimizer = OptimizerType.create_optimizer(
            # pyre-fixme[16]: `SyncServer` has no attribute `cfg`.
            config=self.cfg.server_optimizer,
            model=global_model.fl_get_module(),
        )
        self._global_model: IFLModel = global_model
        self._aggregator: Aggregator = Aggregator(
            module=global_model.fl_get_module(),
            aggregation_type=self.cfg.aggregation_type,
            only_federated_params=self.cfg.only_federated_params,
        )
        self._active_user_selector = instantiate(self.cfg.active_user_selector)
        self._channel: IFLChannel = channel or IdentityChannel()

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        if OmegaConf.is_missing(cfg.active_user_selector, "_target_"):
            cfg.active_user_selector = UniformlyRandomActiveUserSelectorConfig()
        if OmegaConf.is_missing(cfg.server_optimizer, "_target_"):
            cfg.server_optimizer = FedAvgOptimizerConfig()

    @property
    def global_model(self):
        return self._global_model

    def select_clients_for_training(
        self,
        num_total_users,
        users_per_round,
        data_provider: Optional[IFLDataProvider] = None,
        epoch: Optional[int] = None,
    ):
        return self._active_user_selector.get_user_indices(
            num_total_users=num_total_users,
            users_per_round=users_per_round,
            data_provider=data_provider,
            global_model=self.global_model,
            epoch=epoch,
        )

    def init_round(self):
        self._aggregator.zero_weights()
        self._optimizer.zero_grad()

    def receive_update_from_client(self, message: Message):
        message = self._channel.client_to_server(message)

        self._aggregator.apply_weight_to_update(
            delta=message.model.fl_get_module(), weight=message.weight
        )
        self._aggregator.add_update(
            delta=message.model.fl_get_module(), weight=message.weight
        )

    def step(self):
        aggregated_model = self._aggregator.aggregate()
        FLModelParamUtils.set_gradient(
            model=self._global_model.fl_get_module(),
            reference_gradient=aggregated_model,
        )
        self._optimizer.step()


@dataclass
class SyncServerConfig:
    _target_: str = fullclassname(SyncServer)
    _recursive_: bool = False
    only_federated_params: bool = True
    aggregation_type: AggregationType = AggregationType.WEIGHTED_AVERAGE
    server_optimizer: ServerOptimizerConfig = ServerOptimizerConfig()
    active_user_selector: ActiveUserSelectorConfig = ActiveUserSelectorConfig()



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from __future__ import annotations

from collections import OrderedDict
from dataclasses import dataclass

import torch
from flsim.channels.base_channel import (
    IdentityChannel,
    FLChannelConfig,
    Message,
)
from flsim.utils.config_utils import fullclassname
from flsim.utils.config_utils import init_self_cfg
from torch.quantization.observer import (
    MinMaxObserver,
    PerChannelMinMaxObserver,
)


class ScalarQuantizationChannel(IdentityChannel):
    """
    Implements a channel that emulates scalar quantization from 1 to 8
    bits per weight (8 bits per weight corresponds to int8 quantization).
    We emulate this by successively quantizing and dequantizing. This way,
    the rest of the training is transparent for aggreagators, reducers,
    trainers and so on.

    Notes:
        - We can perform either per_tensor quantization (same scale and
          zero_point for every parameters in a weight matrix) or per_channel
          quantization (each channel has its own scale and zero_point). Set
          quantize_per_tensor = False to perform per_channel quantization.
        - We rely on the very simple MinMax observers for both per_tensor
          and per_channel quantization. This can be refined by leveraging the
          HistogramObserver for instance.
        - We do not quantize the biases for the moment since their compression
          overhead is very small.
        - We arbitrarily choose to set the int_repr() of a quantized tensor
          to [0, 2 ** n_bits - 1]. This is called unsigned quantization.
        - All the quantized tensors share the same type, ie `torch.quint8`.
          However, when quantizing to less than 8 bits, this is not memory
          efficient since each element is stored over 1 byte anyway. Since
          we are interested only in emulation for the moment, that's good.
          We could also have relied on the fake_quantize primitives but we
          prefer to rely on the true quantization operators.
    """

    def __init__(self, **kwargs):
        init_self_cfg(
            self,
            component_class=__class__,
            config_class=ScalarQuantizationChannelConfig,
            **kwargs,
        )
        super().__init__(**kwargs)
        self.n_bits = self.cfg.n_bits
        self.quantize_per_tensor = self.cfg.quantize_per_tensor
        assert (
            0 < self.n_bits <= 8
        ), "ScalarQuantizationChannel expects n_bits between 1 and 8 (included)."
        self.quant_min = 0
        self.quant_max = 2 ** self.n_bits - 1
        self.observer, self.quantizer = self._get_observers_and_quantizers()

    def _calc_message_size_client_to_server(self, message: Message):
        """
        We compute the size of the compressed message as follows:
            - for the weights (compressed): n_bits / 8 bytes per element
            - for the biases (not compressed): 4 bytes per element
            - for the scales (one for each layer or one for each layer channel
              depending on quantize_per_tensor): 8 bytes / element (fp64)
            - for the zerp_points (one for each layer or one for each layer channel
              depending on quantize_per_tensor): 4 bytes / element (int32)
        """
        message_size_bytes = 0
        for param in message.model_state_dict.values():
            if param.ndim > 1:
                # integer representation
                message_size_bytes += param.numel() * self.n_bits / 8
                # size of scale(s) (fp64) and zero_point(s) (int32)
                if self.quantize_per_tensor:
                    message_size_bytes += ScalarQuantizationChannel.BYTES_PER_FP64
                    message_size_bytes += ScalarQuantizationChannel.BYTES_PER_FP32
                else:
                    # pyre-ignore[16]: `torch.Tensor` has no attribute `q_per_channel_scales`
                    n_scales = param.q_per_channel_scales().numel()
                    # pyre-ignore[16]: `torch.Tensor` has no attribute `q_per_channel_zero_points`
                    n_zero_points = param.q_per_channel_zero_points().numel()
                    message_size_bytes += (
                        ScalarQuantizationChannel.BYTES_PER_FP64 * n_scales
                    )
                    message_size_bytes += (
                        ScalarQuantizationChannel.BYTES_PER_FP32 * n_zero_points
                    )
            else:
                message_size_bytes += 4 * param.numel()
        return message_size_bytes

    def _get_observers_and_quantizers(self):
        if self.quantize_per_tensor:
            observer = MinMaxObserver(
                dtype=torch.quint8,
                qscheme=torch.per_tensor_affine,
                quant_min=self.quant_min,
                quant_max=self.quant_max,
                reduce_range=False,
            )
            quantizer = torch.quantize_per_tensor
        else:
            observer = PerChannelMinMaxObserver(
                dtype=torch.quint8,
                qscheme=torch.per_channel_affine,
                quant_min=self.quant_min,
                quant_max=self.quant_max,
                reduce_range=False,
                ch_axis=0,
            )
            quantizer = torch.quantize_per_channel
        return observer, quantizer

    def _quantize(self, x: torch.Tensor) -> torch.Tensor:

        # important to reset values, otherwise takes running min and max
        self.observer.reset_min_max_vals()

        # forward through the observer to get scale(s) and zero_point(s)
        _ = self.observer(x)
        scale, zero_point = self.observer.calculate_qparams()

        # Emulate quantization. Not a no-op since we lose precision when quantizing.
        if self.quantize_per_tensor:
            xq = self.quantizer(x, float(scale), int(zero_point), dtype=torch.quint8)
        else:
            scale = scale.to(x.device)
            zero_point = zero_point.to(x.device)
            xq = self.quantizer(x, scale, zero_point, axis=0, dtype=torch.quint8)
        return xq

    @classmethod
    def _set_defaults_in_cfg(cls, cfg):
        pass

    def _on_client_before_transmission(self, message: Message) -> Message:
        """
        We quantize the weights but do not quantize the biases since
        the overhead is very small. We copy the state dict since the
        tensor format changes.
        """
        message.populate_state_dict()
        new_state_dict = OrderedDict()
        for name, param in message.model_state_dict.items():
            if param.ndim > 1:
                new_state_dict[name] = self._quantize(param.data)
            else:
                new_state_dict[name] = param.data

        message.model_state_dict = new_state_dict
        return message

    def _on_server_before_transmission(self, message: Message) -> Message:
        message.populate_state_dict()
        return message

    def _on_server_after_reception(self, message: Message) -> Message:
        """
        We dequantize the weights and do not dequantize the biases
        since they have not beed quantized in the first place. We
        copy the state dict since the tensor format changes.
        """
        new_state_dict = OrderedDict()
        for name, param in message.model_state_dict.items():
            if param.ndim > 1:
                # pyre-ignore[16]: `torch.Tensor` has no attribute `dequantize`
                new_state_dict[name] = param.data.dequantize()
            else:
                new_state_dict[name] = param.data
        message.model_state_dict = new_state_dict
        message.update_model_()
        return message


@dataclass
class ScalarQuantizationChannelConfig(FLChannelConfig):
    _target_: str = fullclassname(ScalarQuantizationChannel)
    n_bits: int = 8
    quantize_per_tensor: bool = True



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from typing import List, Type

import pytest
from flsim.clients.async_client import AsyncClientDevice, AsyncClientFactory
from flsim.clients.base_client import ClientConfig
from flsim.common.pytest_helper import assertEqual, assertIsNotNone
from flsim.data.data_provider import FLDataProviderFromList
from flsim.utils.async_trainer.async_user_selector import (
    RandomAsyncUserSelector,
    RoundRobinAsyncUserSelector,
)
from flsim.utils.async_trainer.device_state import TrainingState
from flsim.utils.async_trainer.training_event_generator import (
    AsyncTrainingEventGenerator,
    AsyncTrainingEventGeneratorConfig,
    AsyncTrainingEventGeneratorFromList,
    AsyncTrainingEventGeneratorFromListConfig,
    EventTimingInfo,
    PoissonAsyncTrainingStartTimeDistrConfig,
)
from flsim.utils.data.fake_data_utils import create_mock_data_provider
from flsim.utils.sample_model import DummyAlphabetFLModel, MockFLModel
from flsim.utils.tests.helpers.test_data_utils import DummyAlphabetDataset
from flsim.utils.timing.training_duration_distribution import (
    PerExampleGaussianDurationDistributionConfig,
    PerUserGaussianDurationDistributionConfig,
    DurationDistributionConfig,
)
from omegaconf import OmegaConf


@pytest.fixture(scope="class")
def prepare_shared_client_config(request):
    request.cls.shared_client_config = OmegaConf.structured(
        ClientConfig(
            epochs=1,
            max_clip_norm_normalized=0,
            only_federated_params=True,
            random_seed=1,
            store_models_and_optimizers=False,
        )
    )


@pytest.mark.usefixtures("prepare_shared_client_config")
class TestAsyncClientDeviceGeneration:
    def _verify_event(
        self,
        client: AsyncClientDevice,
        expected_start_time: int,
        expected_end_time: int,
    ):
        assertEqual(client.training_schedule.start_time, expected_start_time)
        assertEqual(client.training_schedule.end_time, expected_end_time)

    def test_provide_client_event_generation(self) -> None:
        r"""
        Check if client provider returns the client with the correct
        start time and end time
        """
        # (start time, duration)
        event_list = [
            EventTimingInfo(prev_event_start_to_current_start=1, duration=3),
            EventTimingInfo(prev_event_start_to_current_start=2, duration=5),
            EventTimingInfo(prev_event_start_to_current_start=2, duration=1),
            EventTimingInfo(prev_event_start_to_current_start=10, duration=10),
        ]
        start_times_gaps = [val.prev_event_start_to_current_start for val in event_list]
        start_times = [
            sum(start_times_gaps[0 : (x + 1)]) for x in range(0, len(start_times_gaps))
        ]
        durations = [d.duration for d in event_list]
        end_times = [t[0] + t[1] for t in zip(start_times, durations)]
        event_generator = AsyncTrainingEventGeneratorFromList(
            **OmegaConf.structured(
                AsyncTrainingEventGeneratorFromListConfig(training_events=event_list)
            )
        )
        num_users = len(event_list)
        data_provider = create_mock_data_provider(
            num_users=num_users, examples_per_user=1
        )
        user_selector = RandomAsyncUserSelector(data_provider=data_provider)

        current_time = 0
        for start, end in zip(start_times, end_times):
            client = AsyncClientFactory.create(
                current_time=current_time,
                event_generator=event_generator,
                user_selector=user_selector,
                # pyre-ignore [16]: for pytest
                client_config=self.shared_client_config,
            )
            self._verify_event(client, expected_start_time=start, expected_end_time=end)
            # how we move forward in time in async is by setting the current time
            # to start time of the client on top of the heap
            current_time = client.next_event_time()

    def test_sequential_client_training_schedule(self) -> None:
        r"""
        Check that training event generator produces TrainingSchedule sequentially
        (where mean and SD of training time is 0), clients are truly produced sequentialy:
        i.e, if client starts training, client A ends training before any other client start
        """
        num_users = 100
        training_start_time_distr = PoissonAsyncTrainingStartTimeDistrConfig(
            training_rate=10
        )
        duration_distr = PerExampleGaussianDurationDistributionConfig(
            training_duration_mean=0, training_duration_sd=0
        )
        event_generator = AsyncTrainingEventGenerator(
            **OmegaConf.structured(
                AsyncTrainingEventGeneratorConfig(
                    training_start_time_distribution=training_start_time_distr,
                    duration_distribution_generator=duration_distr,
                )
            )
        )
        examples_per_user = 1
        data_provider = create_mock_data_provider(
            num_users=num_users, examples_per_user=examples_per_user
        )
        user_selector = RandomAsyncUserSelector(data_provider=data_provider)

        current_time = 0
        clients = []
        for _ in range(num_users):
            client = AsyncClientFactory.create(
                current_time=current_time,
                event_generator=event_generator,
                user_selector=user_selector,
                # pyre-ignore [16]: for pytest
                client_config=self.shared_client_config,
            )
            assertEqual(
                client.training_schedule.start_time, client.training_schedule.end_time
            )
            current_time = client.next_event_time()
            clients.append(client)

        # verify that clients were produced and hence trained sequentially
        for client_1, client_2 in zip(clients, clients[1:]):
            # check that client_1 should end trainign before  client_2 start training
            assert (
                client_1.training_schedule.end_time
                <= client_2.training_schedule.start_time
            )

            # check that start time is strictly monotonic increasing
            assert (
                client_1.training_schedule.start_time
                < client_2.training_schedule.start_time
            )

    def _build_clients_training_duration_dist(
        self, duration_distr_config: Type[DurationDistributionConfig], num_users: int
    ) -> List[AsyncClientDevice]:
        r"""
        Per-Example-Gaussian:
            training_duration = num_examples * training_duration_per_example
        Per-User-Gaussian:
            training_duration = training_duration_per_example

        Use a config where training time is completely determined by
        number of examples. Eg:
            - GaussianDuration,PoissonStartTime
            - training_rate: very high
            - mean training time: very high
        num_examples_per_user = [n, n-1, n-2, ..., 3, 2, 1]
        """
        training_start_time_distr = PoissonAsyncTrainingStartTimeDistrConfig(
            training_rate=1000
        )
        duration_distr = duration_distr_config(
            training_duration_mean=1000,
            training_duration_sd=0,
        )
        event_generator = AsyncTrainingEventGenerator(
            **OmegaConf.structured(
                AsyncTrainingEventGeneratorConfig(
                    training_start_time_distribution=training_start_time_distr,
                    duration_distribution_generator=duration_distr,
                )
            )
        )
        num_examples_per_user = list(reversed(range(1, num_users + 1)))
        data = [
            [1] * num_example
            for num_example, _ in zip(num_examples_per_user, range(num_users))
        ]
        data_provider = FLDataProviderFromList(
            train_user_list=data,
            eval_user_list=data,
            test_user_list=data,
            model=MockFLModel(),
        )
        user_selector = RoundRobinAsyncUserSelector(data_provider=data_provider)

        clients = []
        current_time = 0
        for _ in range(num_users):
            client = AsyncClientFactory.create(
                current_time=current_time,
                event_generator=event_generator,
                user_selector=user_selector,
                # pyre-ignore [16]: for pytest
                client_config=self.shared_client_config,
            )
            current_time = client.next_event_time()
            clients.append(client)
        return clients

    def test_training_duration_per_example_gaussian(self):
        r"""
        Per-Example-Gaussian:
            The last user should finish first.
            So training finish time should be: [user n, user n-1, ..., user2, user1]
        """
        num_users = 50
        clients = self._build_clients_training_duration_dist(
            duration_distr_config=PerExampleGaussianDurationDistributionConfig,
            num_users=num_users,
        )
        # check that end time is strictly monotonic decreasing
        for client_1, client_2 in zip(clients, clients[1:]):
            assert (
                client_1.training_schedule.end_time
                > client_2.training_schedule.end_time
            )

    def test_training_duration_per_user_gaussian(self):
        r"""
        Per-User-Gaussian:
            First user should finish first
            So training finish time should be: [user 1, user 2, user 3, .... user n]
        """
        num_users = 50
        clients = self._build_clients_training_duration_dist(
            duration_distr_config=PerUserGaussianDurationDistributionConfig,
            num_users=num_users,
        )
        # check that end time is strictly monotonic increasing
        for client_1, client_2 in zip(clients, clients[1:]):
            assert (
                client_1.training_schedule.end_time
                < client_2.training_schedule.end_time
            )


@pytest.fixture(scope="class")
def prepare_async_client_device(request):
    request.cls.shared_client_config = OmegaConf.structured(
        ClientConfig(
            epochs=1,
            max_clip_norm_normalized=0,
            only_federated_params=True,
            random_seed=1,
            store_models_and_optimizers=False,
        )
    )
    request.cls.event_list = [
        EventTimingInfo(prev_event_start_to_current_start=1, duration=3),
        EventTimingInfo(prev_event_start_to_current_start=2, duration=5),
        EventTimingInfo(prev_event_start_to_current_start=2, duration=1),
        EventTimingInfo(prev_event_start_to_current_start=10, duration=10),
    ]
    request.cls.event_generator = AsyncTrainingEventGeneratorFromList(
        **OmegaConf.structured(
            AsyncTrainingEventGeneratorFromListConfig(
                training_events=request.cls.event_list
            )
        )
    )


@pytest.mark.usefixtures("prepare_async_client_device")
class TestAsyncClientDevice:
    def _build_data_provider(
        self, num_examples, examples_per_user, user_batch_size, global_model
    ) -> FLDataProviderFromList:
        dummy_dataset = DummyAlphabetDataset(num_examples)
        data_provider, _ = DummyAlphabetDataset.create_data_provider_and_loader(
            dummy_dataset, examples_per_user, user_batch_size, global_model
        )
        return data_provider

    def test_async_client_start_training(self) -> None:
        global_model = DummyAlphabetFLModel()

        examples_per_user = 10
        data_provider = self._build_data_provider(
            num_examples=100,
            examples_per_user=examples_per_user,
            user_batch_size=2,
            global_model=global_model,
        )

        user_selector = RandomAsyncUserSelector(data_provider=data_provider)

        client = AsyncClientFactory.create(
            current_time=0,
            # pyre-ignore [16]: for pytest
            event_generator=self.event_generator,
            user_selector=user_selector,
            # pyre-ignore [16]: for pytest
            client_config=self.shared_client_config,
        )
        current_seqnum = 1
        # verify that client is waiting to start
        assert client.is_waiting_to_start()

        client.training_started(model_seqnum=current_seqnum, init_model=global_model)
        # verify that that we saved a copy of the global model
        assertIsNotNone(client.local_model)
        # verify that client has the correct seq num
        assertEqual(client.model_seqnum, current_seqnum)
        # verify that client state is training
        assert not client.is_waiting_to_start()
        assertEqual(client.training_state, TrainingState.TRAINING)

    def test_async_client_training(self) -> None:
        num_examples = 10
        examples_per_user = 10
        user_batch_size = 2
        training_start_time = 1
        training_duration = 3
        training_end_time = training_start_time + training_duration

        global_model = DummyAlphabetFLModel()
        data_provider = self._build_data_provider(
            num_examples=num_examples,
            examples_per_user=examples_per_user,
            user_batch_size=user_batch_size,
            global_model=global_model,
        )

        event_list = [
            EventTimingInfo(
                prev_event_start_to_current_start=training_start_time,
                duration=training_duration,
            )
        ]
        event_generator = AsyncTrainingEventGeneratorFromList(
            **OmegaConf.structured(
                AsyncTrainingEventGeneratorFromListConfig(training_events=event_list)
            )
        )

        user_selector = RandomAsyncUserSelector(data_provider=data_provider)
        client = AsyncClientFactory.create(
            current_time=0,
            event_generator=event_generator,
            user_selector=user_selector,
            # pyre-ignore [16]: for pytest
            client_config=self.shared_client_config,
        )

        current_seqnum = 1
        assertEqual(client.next_event_time(), training_start_time)
        client.training_started(model_seqnum=current_seqnum, init_model=global_model)
        (
            client_delta,
            final_local_model,
            num_examples_trained,
        ) = client.train_local_model(metric_reporter=None)
        client.training_ended()

        assertEqual(num_examples_trained, examples_per_user)
        assertEqual(client.training_state, TrainingState.TRAINING_FINISHED)
        assertEqual(client.next_event_time(), training_end_time)
        assertEqual(client.model_seqnum, current_seqnum)

    def test_async_client_less_than(self) -> None:
        num_examples = 10
        examples_per_user = 10
        user_batch_size = 2

        global_model = DummyAlphabetFLModel()
        data_provider = self._build_data_provider(
            num_examples=num_examples,
            examples_per_user=examples_per_user,
            user_batch_size=user_batch_size,
            global_model=global_model,
        )
        user_selector = RandomAsyncUserSelector(data_provider=data_provider)
        # two clients
        # client 1 starts training at 1
        # client 2 starts training at 2
        # verify that client 1 will be less than client 2
        event_list = [
            EventTimingInfo(prev_event_start_to_current_start=1, duration=1),
            EventTimingInfo(prev_event_start_to_current_start=2, duration=1),
        ]
        event_generator = AsyncTrainingEventGeneratorFromList(
            **OmegaConf.structured(
                AsyncTrainingEventGeneratorFromListConfig(training_events=event_list)
            )
        )
        client_1 = AsyncClientFactory.create(
            current_time=0,
            event_generator=event_generator,
            user_selector=user_selector,
            # pyre-ignore [16]: for pytest
            client_config=self.shared_client_config,
        )
        client_2 = AsyncClientFactory.create(
            current_time=0,
            event_generator=event_generator,
            user_selector=user_selector,
            client_config=self.shared_client_config,
        )
        assert client_1 < client_2

        # two clients currently training (training_state=TRAINING_STARTED)
        # client a ends training at 2
        # client b ends training at 3
        # verify that client a will be less than client b
        event_list = [
            EventTimingInfo(prev_event_start_to_current_start=1, duration=1),
            EventTimingInfo(prev_event_start_to_current_start=1, duration=2),
        ]
        event_generator = AsyncTrainingEventGeneratorFromList(
            **OmegaConf.structured(
                AsyncTrainingEventGeneratorFromListConfig(training_events=event_list)
            )
        )
        client_a = AsyncClientFactory.create(
            current_time=0,
            event_generator=event_generator,
            user_selector=user_selector,
            client_config=self.shared_client_config,
        )
        client_b = AsyncClientFactory.create(
            current_time=0,
            event_generator=event_generator,
            user_selector=user_selector,
            client_config=self.shared_client_config,
        )
        client_b.training_started(1, global_model)
        assert client_a < client_b



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from hydra.core.config_store import ConfigStore

from .async_trainer import AsyncTrainerConfig
from .private_sync_trainer import PrivateSyncTrainerConfig
from .sync_trainer import SyncTrainerConfig

ConfigStore.instance().store(
    name="base_sync_trainer",
    node=SyncTrainerConfig,
    group="trainer",
)

ConfigStore.instance().store(
    name="base_async_trainer",
    node=AsyncTrainerConfig,
    group="trainer",
)

ConfigStore.instance().store(
    name="base_private_sync_trainer",
    node=PrivateSyncTrainerConfig,
    group="trainer",
)



#!/usr/bin/env python3
# (c) Facebook, Inc. and its affiliates. Confidential and proprietary.

from typing import Any, Callable, Iterator, Iterable, List, Optional

from flsim.data.data_provider import (
    FLDataProviderFromList,
    IFLDataProvider,
    IFLUserData,
)
from flsim.utils.sample_model import MockFLModel


def create_mock_data_provider(
    num_users: int, examples_per_user: int
) -> IFLDataProvider:
    # one_user_data has 1 batch of len = examples_per_user
    one_user_data = [list(range(examples_per_user))]
    data = [one_user_data] * num_users

    return FLDataProviderFromList(
        train_user_list=data,
        eval_user_list=data,
        test_user_list=data,
        model=MockFLModel(num_examples_per_user=examples_per_user),
    )


class FakeUserData(IFLUserData):
    """
    fake data for a single user.
    """

    def __init__(
        self,
        gen_batch: Callable[[int, Any], Any],
        num_batches: int = 1,
        batch_size: int = 2,
        val: Optional[float] = None,
    ):
        """

        gen_batch is a callable that gets a batch_size
        and generates a simulated batch with the same size
        """
        self.gen_batch = gen_batch
        self._num_batches = num_batches
        self.batch_size = batch_size
        self.val = val

    def __iter__(self) -> Iterator[Any]:
        # TODO add flag for a final batch being incomplete
        for _ in range(self._num_batches):
            yield self.gen_batch(self.batch_size, self.val)

    def num_examples(self) -> int:
        return self._num_batches * self.batch_size

    def num_batches(self) -> int:
        return self._num_batches


class FakeDataProvider(IFLDataProvider):
    def __init__(
        self,
        gen_batch: Callable[[int, Any], Any],
        num_batches: int = 1,
        batch_size: int = 2,
        num_users: int = 10,
        random: bool = False,
        rank: int = 0,
        world_size: int = 1,
    ):
        self.user_data = [
            FakeUserData(
                gen_batch, num_batches, batch_size, None if random else i / num_users
            )
            for i in range(rank, num_users, world_size)
        ]
        self.eval = FakeUserData(
            gen_batch, num_batches, batch_size, None if random else 1
        )
        self._num_users = num_users // world_size
        self._num_total_users = num_users

    def __iter__(self) -> Iterable[IFLUserData]:
        yield from self.user_data

    def __getitem__(self, index) -> IFLUserData:
        return self.user_data[index]

    def user_ids(self) -> List[int]:
        return list(range(self._num_users))

    def num_users(self) -> int:
        return self._num_users

    def num_total_users(self) -> int:
        return self._num_total_users

    def get_user_data(self, user_index: int) -> IFLUserData:
        return self[user_index]

    def train_data(self) -> Iterable[IFLUserData]:
        return self.user_data

    def eval_data(self) -> Iterable[Any]:
        return self.eval

    def test_data(self) -> Iterable[Any]:
        return self.eval
